/*
 * Copyright (c) 1998 The University of Bonn
 * All rights reserved.
 * 
 * Permission to use and copy this software in source and binary forms
 * is hereby granted, provided that the above copyright notice, this
 * paragraph and the following disclaimer are retained in any copies
 * of any part of this software and that the University of Bonn is
 * acknowledged in all documentation pertaining to any such copy
 * or derivative work. The name of the University of Bonn may not
 * be used to endorse or promote products derived from this software
 * without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE WARRANTIES OF 
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. IN NO EVENT SHALL 
 * THE UNIVERSITY OF BONN BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF 
 * OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE 
 * SOFTWARE.
 */


#include "rsvp.h"
#include "rsvp-objects.h"
#include "rsvp-messages.h"
#include "rng.h"


int hdr_rsvp::offset_;

static class RSVPHeaderClass : public PacketHeaderClass {
public:
  RSVPHeaderClass() : PacketHeaderClass("PacketHeader/RSVP", 
					sizeof(hdr_rsvp)) 
  {
    bind_offset(&hdr_rsvp::offset_);
  }
} class_RSVPhdr;


static class RSVPClass : public TclClass {
public:
  RSVPClass() : TclClass("Agent/RSVP") {}
  TclObject* create(int, const char*const*) {
    return (new RSVPAgent());
  }
} class_RSVP;


RSVPAgent::RSVPAgent() : Agent(PT_RSVP), noisy_(0), next_sid_(0), 
  ip6_(0), s_list_(0), t_list_(0) {
  bind_bool("noisy_", &noisy_);
  bind("refresh_", &refresh_);
  bind("lifetime_factor_", &lifetime_factor_);
  bind("ip6_", &ip6_);
  bind("num_flows_", &num_flows_);
  bind("num_rsb_", &num_rsb_);
  bind("num_psb_", &num_psb_);
  bind_bool("nam_", &nam_);
  /* At the moment, drand48 is used as random generator. This should
     be changed at some point to use the ns random generators. */
  //srand48(gethrtime());
  // changed to use the default rng - Sean Murphy 4/12/99
}


/* Check if any path states have timed out and should be removed. This
   function returns a '1' if the session was removed in the process
   (because no state blocks were left), otherwise it returns 0. */
char RSVPAgent::check_path(session *s) {
  psb *temp, *search;
  double now = Scheduler::instance().clock();
  while ((s->psb_list != NULL) && 
	 (s->psb_list->phop->get_hop() != -1) &&
	 (now > s->psb_list->timeout)) {
    temp = s->psb_list;
    s->psb_list = s->psb_list->next;
    send_path_tear_message(s, temp);
    if ((noisy_ & UPC_PATHTO) && (temp->timeout != -1)) {
      Tcl& tcl = Tcl::instance();
      tcl.evalf("%s upcall-path-timeout %d %d", name(),  s->sid, 
		temp->sender->get_addr());
    }
    delete_rsbs(s, temp);  /* (see below) */
    clear_psb(temp);
    delete temp;
    num_psb_--;
  }
  if (s->psb_list != NULL) {
    search = s->psb_list;
    while (search->next != NULL) {
      if ((search->next->phop->get_hop() != -1) &&
	  (search->next->timeout < now)) {
	temp = search->next;
	search->next = temp->next;
	if ((noisy_ & UPC_PATHTO) && (temp->timeout != -1)) {
	  Tcl& tcl = Tcl::instance();
	  tcl.evalf("%s upcall-path-timeout %d %d", name(),  s->sid, 
		    temp->sender->get_addr());
	}
	send_path_tear_message(s, temp);
	/* Find all RSBs for this PSB and tear them down and update traffic
	   control accordingly. Important note: No ResvTear messages will 
	   be sent as a result of this. RFC 2205 states that ResvTear 
	   messages should be sent by end systems or as a result of 
	   *reservation* timeouts. However, here we are dealing with 
	   a *path* timeout. RFC 2209 does not mention anything
	   about this, and the ISI RSVP demon does not seem to send
	   any ResvTear messages either when a path state expires. */
	delete_rsbs(s, temp);
	clear_psb(temp);
	delete temp;
	num_psb_--;
      } else {
	search = search->next;
      }
    }
  }
  if (s->psb_list == NULL) {
    s->path_ref = -1;
  }
  if ((s->resv_ref == -1) && (s->path_ref == -1)) {
    remove_session(s);
    return 1;
  }
  return 0;
}


/* Check if any resv states have timed out and should be removed. This
   function returns 1 if the session was removed in the process, otherwise
   it returns 0. */
char RSVPAgent::check_resv(session *s) {
  rsb *temp, *search, *r_list;
  // Tcl& tcl = Tcl::instance(); - unused
  double now = Scheduler::instance().clock();
  r_list = 0;
  while ((s->rsb_list != NULL) && 
	 (s->rsb_list->nhop->get_hop() != -1) &&
	 (s->rsb_list->timeout < now)) {
    temp = s->rsb_list;
    s->rsb_list = s->rsb_list->next;
    update_traffic_control(s, temp);
    if ((noisy_ & UPC_RESVTO) && (temp->timeout != -1) && 
	(temp->nhop->get_hop() != -1)) {
      Tcl& tcl = Tcl::instance();
      tcl.evalf("%s upcall-resv-timeout %d %d", name(),  s->sid, 
		temp->sender->get_addr());
    }
    temp->next = r_list;
    r_list = temp;
    num_rsb_--;
  }
  if (s->rsb_list != NULL) {
    search = s->rsb_list;
    while (search->next != NULL) {
      if ((search->next->nhop->get_hop() != -1) &&
	  (search->next->timeout < now)) {
	temp = search->next;
	search->next = temp->next;
	clear_rsb(temp);
	update_traffic_control(s, temp);
	if ((noisy_ & UPC_RESVTO) && (temp->timeout != -1) && 
	    (temp->nhop->get_hop() != -1)) {
	  Tcl& tcl = Tcl::instance();
	  tcl.evalf("%s upcall-resv-timeout %d %d", name(),  s->sid, 
		    temp->sender->get_addr());
	}
	temp->next = r_list;
	r_list = temp;
	num_rsb_--;
      } else {
	search = search->next;
      }
    }
  }
  send_resv_tear_messages(s, r_list);
  refresh_resv(s, 0);
  while (r_list != NULL) {
    temp = r_list;
    r_list = r_list->next;
    clear_rsb(temp);
    delete temp;
  }
  if (s->rsb_list == NULL) {
    s->resv_ref = -1;
  }
  if ((s->resv_ref == -1) && (s->path_ref == -1)) {
    remove_session(s);
    return 1;
  }
  return 0;
}


/* Make sure that all sub-objects of a PSB are deleted */
void RSVPAgent::clear_psb(psb *p) {
  delete p->sender;
  delete p->s_tspec;
  delete p->phop;
}


/* Make sure that all sub-objects of an RSB are deleted */
void RSVPAgent::clear_rsb(rsb *r) {
  delete r->nhop;
  delete r->fspec;
  delete r->sender;
  delete r->st;
  r->p = NULL;
}


int RSVPAgent::command(int argc, const char*const* argv) {
  Tcl& tcl = Tcl::instance();
  /* session dest fid */
  if (strcmp(argv[1], "session") == 0) {
    if (argc != 4) {
      return (TCL_ERROR);
    }
    if (argv[2][0] == '_') {
      tcl.evalf("%s id", argv[2]);
      tcl.resultf("%d", new_session_sid(atoi(tcl.result()), atoi(argv[3]), 1));
    } else {
      tcl.resultf("%d", new_session_sid(strtol(argv[2], NULL, 0), 
					atoi(argv[3]), 1));
    }
    return (TCL_OK);
  }
  /* confirm session-id */
  if (strcmp(argv[1], "confirm") == 0) {
    if (argc != 3) {
      return (TCL_ERROR);
    }
    session *s = find_session_sid(atoi(argv[2]));
    if (s == NULL) {
      tcl.resultf("%s RSVP: unknown session ID %s", name(), argv[2]);
      return(TCL_ERROR);
    }
    s->confirm = 1;
    return (TCL_OK);
  }
  /* sender session-id rate bucket-size ttl */
  if (strcmp(argv[1], "sender") == 0) {
    if (argc != 6) {
      return (TCL_ERROR);
    }
    session *s = find_session_sid(atoi(argv[2]));
    if (s == NULL) {
      tcl.resultf("%s RSVP: unknown session ID %s", name(), argv[2]);
      return(TCL_ERROR);
    }
    psb *p;
    // if ((p = find_psb(s, addr_ >> Address::instance().NodeShift_[1])) 
    if ((p = find_psb(s, here_.addr_ )) // SM
	== NULL) { 
      new_psb(s, atof(argv[3]), atoi(argv[4]), atoi(argv[5]), 
	      // addr_ >> Address::instance().NodeShift_[1], -1,
	      here_.addr_ , -1, // SM
	      refresh_, 0, NULL);
      num_psb_++;
    } else {
      update_psb(s, p, atof(argv[3]), atoi(argv[4]), atoi(argv[5]),
		 // addr_ >> Address::instance().NodeShift_[1], -1,
		 here_.addr_ , -1, // SM
		 refresh_, 0, NULL);
    }
    return (TCL_OK);
  }
  /* reserve session-id style <flow descriptor list> */
  if (strcmp(argv[1], "reserve") == 0) {
    if (argc < 6) {
      return (TCL_ERROR);
    }
    session *s = find_session_sid(atoi(argv[2]));
    if (s == NULL) {
      if (noisy_ & UPC_RESVERR) {
	tcl.evalf("%s upcall-resv-error %s 4 0 %d", name(), argv[2],
		  // addr_ >> Address::instance().NodeShift_[1]);
		  here_.addr_ );
      }
    } else {
      /* Are there any PSBs for this session? */
      if (s->psb_list == NULL) {
	if (noisy_ & UPC_RESVERR) {
	  tcl.evalf("%s upcall-resv-error %s 3 0 %d", name(), argv[2],
		    // addr_ >> Address::instance().NodeShift_[1]);
		    here_.addr_ );
	}
      } else {
	if (strcasecmp(argv[3], "ff") != 0) {
	  if (noisy_ & UPC_RESVERR) {
	    tcl.evalf("%s upcall-resv-error %s 6 0 %d", name(), argv[2],
		      // addr_ >> Address::instance().NodeShift_[1]);
		      here_.addr_ );
	  }
	} else {
	  process_ff_request(s, argc, argv);
	}
      }
    }
    return (TCL_OK);
  }
  /* release session-id */
  if (strcmp(argv[1], "release") == 0) {
    if (argc != 3) {
      return (TCL_ERROR);
    }
    session *s = find_session_sid(atoi(argv[2]));
    if (s == NULL) {
      tcl.evalf("puts \"%s RSVP: unknown session ID %s\"", name(), argv[2]);
    } else {   
      release_session(s);
    }
    return (TCL_OK);
  }
  if (strcmp(argv[1], "refresh") == 0) {
    refresh();
    return (TCL_OK);
  }
  /* sessions */
  if (strcmp(argv[1], "sessions") == 0) {
    if (argc != 2) {
      return (TCL_ERROR);
    }
    char buf[200];
    list_sessions(buf, 200);
    tcl.resultf("%s", buf);
    return (TCL_OK);
  }
  /* set-status session-id value */
  if (strcmp(argv[1], "set-status") == 0) {
    if (argc != 4) {
      return (TCL_ERROR);
    }
    session *s = find_session_sid(atoi(argv[2]));
    if (s != NULL) {
      s->status = atoi(argv[3]);
    }
    return (TCL_OK);
  }
  /* get-status session-id */
  if (strcmp(argv[1], "get-status") == 0) {
    if (argc != 3) {
      return (TCL_ERROR);
    }
    session *s = find_session_sid(atoi(argv[2]));
    if (s != NULL) {
      tcl.resultf("%d", s->status);
    }
    return (TCL_OK);
  }
  /* set-handle session-id object */
  if (strcmp(argv[1], "set-handle") == 0) {
    if (argc != 4) {
      return (TCL_ERROR);
    }
    session *s = find_session_sid(atoi(argv[2]));
    if (s != NULL) {
      strcpy(s->handle, argv[3]);
    }
    return (TCL_OK);
  }
  /* get-handle session-id */
  if (strcmp(argv[1], "get-handle") == 0) {
    if (argc != 3) {
      return (TCL_ERROR);
    }
    session *s = find_session_sid(atoi(argv[2]));
    if (s != NULL) {
      tcl.resultf("%s", s->handle);
    }
    return (TCL_OK);
  }
  return (Agent::command(argc, argv));
}


/* Delete all RSBs for the PSB 'p'. */
void RSVPAgent::delete_rsbs(session *s, psb *p) {
  rsb *r, *temp;
  // Tcl& tcl = Tcl::instance(); - unused
  while ((s->rsb_list != NULL) && (s->rsb_list->p == p)) {
    temp = s->rsb_list;
    s->rsb_list = temp->next;
    update_traffic_control(s, temp);
    clear_rsb(temp);
    delete temp;
    num_rsb_--;
  }
  if (s->rsb_list != NULL) {
    r = s->rsb_list;
    while (r->next != NULL) {
      if (r->next->p == p) {
	temp = r->next;
	r->next = temp->next;
	update_traffic_control(s, temp);
	clear_rsb(temp);
	delete temp;
	num_rsb_--;
      } else {
	r = r->next;
      }
    }
  }
  if (s->rsb_list == NULL) {
    s->resv_ref = -1;
  }
}


/* Find a PSB for a session whose SENDER_TEMPLATE matches 'sender'.
   Return NULL if no such PSB is found. */
psb *RSVPAgent::find_psb(session *s, nsaddr_t sender) {
  psb *search = s->psb_list;
  while ((search != NULL) && (search->sender->get_addr() != sender)) {
    search = search->next;
  }
  return search;
}


/* Find an RSB for a session whose FILTERSPEC matches 'sender' and whose
   NHOP matches 'nhop'. Return NULL if no such PSB is found. */
rsb *RSVPAgent::find_rsb(session *s, nsaddr_t sender, nsaddr_t nhop) {
  rsb *search = s->rsb_list;
  while ((search != NULL) && ((search->sender->get_addr() != sender) ||
	 (search->nhop->get_hop() != nhop))) {
    search = search->next;
  }
  return search;
}


/* Return 1 if there is an RSB with the corresponding PSB 'p', otherwise
   return 0. */
int RSVPAgent::find_rsb_for_psb(session *s, psb *p) {
  rsb *r = s->rsb_list;
  while ((r != NULL) && (r->p != p)) {
    r = r->next;
  }
  return (r != NULL);
}


/* Return a pointer to the session with the destination address 'dst'
   and the flow id 'fid' or return NULL if none is found. */
session *RSVPAgent::find_session_dst(nsaddr_t dst, int fid) {
  session *s = s_list_;
  while ((s != NULL) && 
	 ((s->s->get_dest() != dst) || (s->s->get_fid() != fid))) {
    s = s->next;
  }
  if (s == NULL) {
    return NULL;
  } else {
    return s;
  }
}


/* Return a pointer to the session with the id 'sid' or return NULL if
   not found. 
   */
session *RSVPAgent::find_session_sid(long sid) {
  session *s = s_list_;
  while ((s != NULL) && (s->sid > sid)) {
    s = s->next;
  }
  if ((s == NULL) || (s->sid != sid)) {
    return NULL;
  } else {
    return s;
  }
}


void RSVPAgent::give(Packet *p, RSVPChecker *ret) {
  hdr_rsvp* rsvp_hdr = hdr_rsvp::access(p);
  hdr_cmn* cmn_hdr = hdr_cmn::access(p);
  RSVPmessage *msg = new RSVPmessage(p->accessdata());
  switch (msg->get_type()) {
    case PATH: process_path_message(msg, cmn_hdr->iface_, ret); break;
    case PATHTEAR: process_path_tear_message(msg, cmn_hdr->iface_, ret); break;
    case RESV: process_resv_message(msg, rsvp_hdr->fromhop); break;
    case RESVTEAR: process_resv_tear_message(msg); break;
    case RESVERR: process_resv_err_message(msg); break;
    case RESVCONF: process_resv_conf_message(msg); break;
  }
  delete msg;
  Packet::free(p);
}


/* This function is needed to correct the problem that there seem to
   be some accuracy problems in C++ so that two doubles are not necessarily
   equal even if they 'look' equal. The range of 0.00001 seconds should
   not affect the simulations at all (where the refresh time is chosen
   randomly from a certain range anyway). */
char RSVPAgent::in_range(double x, double y) {
  return (x - 0.00001) <= y;
}


/* Checks for a packet if the current node is a leaf node in a multicast
   tree. */
char RSVPAgent::is_leaf(nsaddr_t src, nsaddr_t dst) {
  Tcl& tcl = Tcl::instance();
  tcl.evalf("%s is-leaf %d %d", name(), src, dst);
  return (atoi(tcl.result()));
}


void RSVPAgent::list_sessions(char *buf, int size) {
  session *s = s_list_;
  char *ptr;
  ptr = buf;
  buf[0] = '\0';
  while ((s != NULL) && (ptr < buf + size - 10)) {
    sprintf(ptr, "%ld ", s->sid);
    s = s->next;
    ptr = ptr + strlen(ptr);
  }
}


/* Returns the path_ref or resv_ref of a session, whichever is lesser */
double RSVPAgent::minref(session *s) {
  if (s->path_ref == -1) {
    return s->resv_ref;
  }
  if (s->resv_ref == -1) {
    return s->path_ref;
  }
  if (s->path_ref < s->resv_ref) {
    return s->path_ref;
  } else {
    return s->resv_ref;
  }
}


/* Create a new PSB for a given session, add it to the session's PSB list,
   reschedule the session for refreshing.
   */
void RSVPAgent::new_psb(session *s, double rate, long bucket, char ttl,
			nsaddr_t sender, nsaddr_t phop, double refresh,
			int iface, RSVPChecker *check) {
  psb *p = new psb();
  p->sender = new SENDER_TEMPLATE(sender, ip6_);
  p->s_tspec = new SENDER_TSPEC(rate, bucket);
  p->phop = new RSVP_HOP(phop, ip6_);
  p->ttl = ttl;
  p->iif = iface;
  p->check = check;
  if (phop > -1) { // Call came from API
    p->timeout = (lifetime_factor_ + 0.5) * 1.5 * refresh + 
      Scheduler::instance().clock();
  }
  if ((s->psb_list == NULL) || (s->psb_list->sender->get_addr() > sender)) {
    p->next = s->psb_list;
    s->psb_list = p;
  } else {
    psb *search = s->psb_list;
    while ((search->next != NULL) && 
	   (search->next->sender->get_addr() < sender)) {
      search = search->next;
    }
    p->next = search->next;
    search->next = p;
  }
  send_path_message(s, p);
  /* Reschedule the session if this is the first PSB. Send a path
     message for this PSB. */
  if (s->path_ref == -1) {
    s->path_ref = Scheduler::instance().clock() + s->path_tv->get_r();
    reschedule_session(s);
  }
  /* Make an upcall to the API if necessary */
  if ((noisy_ & UPC_PATH) && (phop > -1)) {
    Tcl& tcl = Tcl::instance();
    tcl.evalf("%s upcall-path %d %f %d %d", name(),  s->sid, rate, 
	      bucket, sender);
  }
}


/* Create a new RSB for a given session and add it to the session's 
   RSB list. */
rsb *RSVPAgent::new_rsb(session *s, psb *p, nsaddr_t sender, FLOWSPEC *fl,
			const char *style, nsaddr_t nhop, double refresh,
			nsaddr_t fromhop) {
  rsb *r = new rsb();
  r->nhop = new RSVP_HOP(nhop, ip6_);
  r->sender = new FILTER_SPEC(sender, ip6_);
  r->fspec = new FLOWSPEC(fl->get_contents());
  r->p = p;
  r->oifhop = fromhop;
  r->modified = 1;
  r->is_new = 1;
  /* Only valid values for 'style' should arrive here */
  if (strcasecmp(style, "FF") == 0) {
    r->st = new STYLE(STYLE_FF);
  }
  if (nhop > -1) { // Call came from API
    r->timeout = (lifetime_factor_ + 0.5) * 1.5 * refresh + 
      Scheduler::instance().clock();
  }
  if ((s->rsb_list == NULL) || (s->rsb_list->sender->get_addr() > sender)) {
    s->rsb_list = r;
  } else {
    rsb *search = s->rsb_list;
    while ((search->next != NULL) && 
	   (search->next->sender->get_addr() < sender)) {
      search = search->next;
    }
    r->next = search->next;
    search->next = r;
  }
  return r;
}


/* Create a new session and add it to the session list, then return
   a pointer to the session. */
session *RSVPAgent::new_session_p(nsaddr_t dst, int fid, char local) {

  RNG* defaultRNG = RNG::defaultrng() ;
  
  session *n, *s;
  n = new session();
  n->s = new SESSION(dst, 0, fid, ip6_);
  // Set the refresh timer as described in section 3.7 of RFC 2205:
  n->path_tv = new TIME_VALUES((defaultRNG->uniform_double() + 0.5) *
			       refresh_ );
  n->resv_tv = new TIME_VALUES((defaultRNG->uniform_double() + 0.5) *
			       refresh_ );
  n->sid = next_sid_;
  n->local = local;
  next_sid_++;
  if (next_sid_ > MAXSID) {
    next_sid_ = 0;
  }
  // The session list is sorted in descending order of the sid's
  if ((s_list_ == NULL) || (s_list_->sid < n->sid)) {
    n->next = s_list_;
    s_list_ = n;
  } else {
    s = s_list_;
    while ((s->next != NULL) && (s->next->sid > n->sid)) {
      s = s->next;
    }
    n->next = s->next;
    s->next = n;
  }
  return n;
}


/* Create a new session and add it to the session list, then return the
   sid. */
int RSVPAgent::new_session_sid(nsaddr_t dst, int fid, char local) {

  RNG* defaultRNG = RNG::defaultrng() ;

  session *n, *s;
  n = new session();
  n->s = new SESSION(dst, 0, fid, ip6_);
  // Set the refresh timer as described in section 3.7 of RFC 2205:
  n->path_tv = new TIME_VALUES((defaultRNG->uniform_double() + 0.5) *
			       refresh_ );
  n->resv_tv = new TIME_VALUES((defaultRNG->uniform_double() + 0.5) *
			       refresh_ );
  n->sid = next_sid_;
  n->local = local;
  next_sid_++;
  if (next_sid_ > MAXSID) {
    next_sid_ = 0;
  }
  // The session list is sorted in descending order of the sid's
  if ((s_list_ == NULL) || (s_list_->sid < n->sid)) {
    n->next = s_list_;
    s_list_ = n;
  } else {
    s = s_list_;
    while ((s->next != NULL) && (s->next->sid > n->sid)) {
      s = s->next;
    }
    n->next = s->next;
    s->next = n;
  }
  return n->sid;
}


void RSVPAgent::process_ff_request(session *s, int argc, 
				  const char*const* argv) {
  Tcl &tcl = Tcl::instance();
  int i = 4;
  rsb *r;
  psb *p;
  nsaddr_t sender;
  char modified = 0;
  FLOWSPEC *fl;
  /* The following code parses a flow descriptor list which matches 
     the following Backus-Naur form:
     <flow descriptor list> ::= <FLOWSPEC> <FILTER_SPEC> |
     <flow descriptor list><FF flow descriptor>
     <FF flow descriptor> ::= [ <FLOWSPEC> ] <FILTER_SPEC>
     This conforms with the description in RFC 2205 and means that there 
     can be any number of FILTER_SPEC objects for a FLOWSPEC object. */
  while (i < argc) {
    if (i + 1 >= argc) {
      tcl.evalf("%s RSVP: FLOWSPEC incomplete", name());
    }
    fl = new FLOWSPEC(atoi(argv[i]), atoi(argv[i+1]));
    i += 2;
    if ((i >= argc) || (argv[i][0] == '+')) {
      tcl.evalf("%s RSVP: FILTER_SPEC missing", name());
    }
    while ((i < argc) && (argv[i][0] != '+')) {
      if (argv[i][0] == '_') {
	tcl.evalf("%s id", argv[i]);
	sender = atoi(tcl.result());
      } else {
	sender = atoi(argv[i]);
      }
      if (((p = find_psb(s, sender)) == NULL) && (noisy_ & UPC_RESVERR)) {
       	tcl.evalf("%s upcall-resv-error %s 4 0 %d %d", name(), argv[2],
		  // addr_ >> Address::instance().NodeShift_[1], sender);
		  here_.addr_ , sender);
      }
      if ((r = find_rsb(s, sender, -1)) == NULL) {
	new_rsb(s, p, sender, fl, "FF", -1, refresh_, -1);
	num_rsb_++;
      } else {
	update_rsb(s, r, p, sender, fl, "FF", -1, refresh_, -1);
      }
      i++;
    }
    delete fl;
  }
  rsb *rsearch = s->rsb_list;
  while (rsearch != NULL) {
    if (rsearch->modified || rsearch->is_new) {
      modified = 1;
      rsearch->modified = 0;
      rsearch->is_new = 0;
    }
    rsearch = rsearch->next;
  }
  refresh_resv(s, 0);
}


void RSVPAgent::process_path_message(RSVPmessage *msg, int iface,
				     RSVPChecker *check) {
  SESSION *se = (SESSION *)msg->get_first(1);
  session *s;
  if ((s = find_session_dst(se->get_dest(), se->get_fid())) == NULL) {
    s = new_session_p(se->get_dest(), se->get_fid(), 0);
  };
  psb *p;
  SENDER_TEMPLATE *sender = (SENDER_TEMPLATE *)msg->get_first(11);
  TIME_VALUES *tv = (TIME_VALUES *)msg->get_first(5);
  SENDER_TSPEC *s_tspec = (SENDER_TSPEC *)msg->get_first(12);
  RSVP_HOP *hop = (RSVP_HOP *)msg->get_first(3);
  if ((p = find_psb(s, sender->get_addr())) == NULL) {
    new_psb(s, s_tspec->get_rate(), s_tspec->get_size(),
	    msg->get_ttl(), sender->get_addr(), hop->get_hop(),
	    tv->get_r(), iface, check);
    num_psb_++;
  } else {
    update_psb(s, p, s_tspec->get_rate(), s_tspec->get_size(),
	       msg->get_ttl(), sender->get_addr(), hop->get_hop(),
	       tv->get_r(), iface, check);
  }
}


void RSVPAgent::process_path_tear_message(RSVPmessage *msg, int iface,
					  RSVPChecker *check) {
  SESSION *se = (SESSION *)msg->get_first(1);
  session *s;
  if ((s = find_session_dst(se->get_dest(), se->get_fid())) != NULL) {
    psb *p;
    SENDER_TEMPLATE *sender = (SENDER_TEMPLATE *)msg->get_first(11);
    RSVP_HOP *hop = (RSVP_HOP *)msg->get_first(3);
    /* Note: Even if there was a SENDER_TSPEC object in this message,
       it is supposed to be ignored according to RFC 2205. */
    if ((p = find_psb(s, sender->get_addr())) != NULL) {
      /* According to RFC 2205, the RSVP_HOP and the PHOP have to be
	 identical. RFC 2209 does not mention that. */
      if (p->phop->get_hop() == hop->get_hop()) {
	if (noisy_ & UPC_PATHTEAR) {
	  Tcl& tcl = Tcl::instance();
	  tcl.evalf("%s upcall-path-tear %d %d", name(), s->sid, 
		    sender->get_addr());
	}
	/* The necessary actions for deleting the path state would
	   be fairly complicated. The PSB  and all corresponding RSBs
	   have to be deleted and the session has to be removed if this 
	   PSB was the last state block. To make this short, a 'trick' 
	   will be used. The timeout for this PSB will be set to -1, and 
	   then a check_path will be performed, which will definitely 
	   remove this PSB (and perhaps even clean up some other PSBs 
	   which are due for removal in the process). */
	p->timeout = -1;
	if (!check_path(s)) {
	  reschedule_session(s);
	}
      }
    }
  }
}


void RSVPAgent::process_resv_conf_message(RSVPmessage *msg) {
  RESV_CONFIRM *confirm = (RESV_CONFIRM *)msg->get_first(15);
  // if (confirm->get_addr() == addr_ >> Address::instance().NodeShift_[1]) {
  if (confirm->get_addr() == here_.addr_ ) {
    SESSION *se = (SESSION *)msg->get_first(1);
    session *s;
    if ((s = find_session_dst(se->get_dest(), se->get_fid())) != NULL) {
      FILTER_SPEC *sender = (FILTER_SPEC *)msg->get_first(10);
      ERROR_SPEC *err = (ERROR_SPEC *)msg->get_first(6);
      if (noisy_ & UPC_RESVCONF) {
	Tcl& tcl = Tcl::instance();
	tcl.evalf("%s upcall-resv-confirm %d %d %d", name(), s->s->get_fid(),
		  sender->get_addr(), err->get_errnode());
      }
    }
  } else {
    if (ip6_) {
      size_ = msg->get_length() + 40; // Size for standard IPv6 header
    } else {
      size_ = msg->get_length() + 24;
    }
    Tcl& tcl = Tcl::instance();
    tcl.evalf("[Simulator instance] connect-rsvp-agents %d %d", 
	      // addr_ >> Address::instance().NodeShift_[1],
	      here_.addr_ ,
	      confirm->get_addr());
    Packet *pkt = allocpkt();
    pkt->allocdata(msg->get_conlength());
    memcpy(pkt->accessdata(), msg->get_contents(), msg->get_conlength());
    send(pkt, 0);
  }
}


void RSVPAgent::process_resv_err_message(RSVPmessage *msg) {
  SESSION *se = (SESSION *)msg->get_first(1);
  session *s;
  rsb *r;
  Tcl& tcl = Tcl::instance();
  if ((s = find_session_dst(se->get_dest(), se->get_fid())) == NULL) {
  } else {
    RSVP_HOP *nhop = (RSVP_HOP *)msg->get_first(3);
    FILTER_SPEC *sender = (FILTER_SPEC *)msg->get_first(10);
    ERROR_SPEC *err = (ERROR_SPEC *)msg->get_first(6);
    /* First mark all RSBs which route to the outgoing interface where
       the ResvErr message came from. */
    while (sender != NULL) {
      r = s->rsb_list;
      while (r != NULL) {
	if ((r->p->phop->get_hop() == nhop->get_hop()) &&
	    (r->sender->get_addr() == sender->get_addr())) {
	  r->processed = 1;
	}
	r = r->next;
      }
      msg->delete_first(10);
      sender = (FILTER_SPEC *)msg->get_first(10);
    }
    /* Now send ResvErr messages to each nhop that appears in the
       marked RSBs. */
    r = s->rsb_list;
    while (r != NULL) {
      if (r->processed) {
	send_resv_err_message(s, msg, r);
      }
      r->processed = 0;
      r = r->next;
    }
    if (noisy_ & UPC_RESVERR) {
      tcl.evalf("%s upcall-resv-error %d %d %d %d", name(), 
		s->sid, err->get_errcode(), err->get_errvalue(), 
		err->get_errnode());
    }
  }
}


void RSVPAgent::process_resv_message(RSVPmessage *msg, nsaddr_t fromhop) {
  SESSION *se = (SESSION *)msg->get_first(1);
  session *s;
  char modified = 0;
  if ((s = find_session_dst(se->get_dest(), se->get_fid())) == NULL) {
    send_resv_err_message(msg, 3);
  } else {
    RSVP_HOP *nhop = (RSVP_HOP *)msg->get_first(3);
    TIME_VALUES *tv = (TIME_VALUES *)msg->get_first(5);
    rsb *r;
    psb *p;
    FLOWSPEC *fl;
    FILTER_SPEC *sender;
    long senderaddr ; // SM
    fl = (FLOWSPEC *)msg->get_first(9);
    while (fl != NULL) {
      sender = (FILTER_SPEC *)msg->get_first(10);
      // There was a problem here - the sender is deleted below with
      // the delete_first method, but is still assumed to contain valid
      // data in the while loop below. To overcome this, I create a
      // new variable - senderaddr - this is used below - SM
      senderaddr = sender->get_addr() ;
      if ((p = find_psb(s, sender->get_addr())) == NULL) {
	send_resv_err_message(msg, 4);
      }
      if ((r = find_rsb(s, sender->get_addr(), nhop->get_hop())) == NULL) {
	r = new_rsb(s, p, sender->get_addr(), fl, "FF", nhop->get_hop(), 
		tv->get_r(), fromhop);
	num_rsb_++;
      } else {
	update_rsb(s, r, p, sender->get_addr(), fl, "FF", nhop->get_hop(), 
		   tv->get_r(), fromhop);
      }
      RESV_CONFIRM *confirm = (RESV_CONFIRM *)msg->get_first(15);
      if (confirm != NULL) {
	r->confirm = new RESV_CONFIRM(confirm->get_contents());
      }
      modified += update_traffic_control(s, r);
      msg->delete_first(9);
      msg->delete_first(10);
      fl = (FLOWSPEC *)msg->get_first(9);
    }
    rsb *rsearch = s->rsb_list;
    while (rsearch != NULL) {
      if (rsearch->modified || rsearch->is_new) {
	rsearch->modified = 0;
	rsearch->is_new = 0;
	/* Make an upcall to the API if necessary. One upcall is made for 
	   each new RSB, and resv upcalls are not only made on leaf nodes. */
	if (noisy_ & UPC_RESV) {
	  Tcl& tcl = Tcl::instance();
	  tcl.evalf("%s upcall-resv %d %f %d %d", name(),  s->sid, 
		    r->fspec->get_rate(), r->fspec->get_size(), 
		    //sender->get_addr());
                    senderaddr) ; // SM
	}
      }
      rsearch = rsearch->next;
    }
    refresh_resv(s, 0);
  }
}


void RSVPAgent::process_resv_tear_message(RSVPmessage *msg) {
  SESSION *se = (SESSION *)msg->get_first(1);
  session *s;
  // Tcl& tcl = Tcl::instance(); - unused
  rsb *r, *r_list, *temp;
  /* Check if a session exists for this ResvTear message, otherwise
     drop the message silently. */
  if ((s = find_session_dst(se->get_dest(), se->get_fid())) != NULL) {
    FILTER_SPEC *sender;
    RSVP_HOP *nhop = (RSVP_HOP *)msg->get_first(3);
    r_list = 0;
    sender = (FILTER_SPEC *)msg->get_first(10);
    /* Read each FILTER_SPEC from the message and delete the corresponding
       RSBs from the rsb_list, putting them in a different list for further
       processing. */
    while (sender != NULL) {
      while ((s->rsb_list != NULL) && 
	     (s->rsb_list->sender->get_addr() == sender->get_addr()) &&
	     (s->rsb_list->nhop->get_hop() == nhop->get_hop())) {
	temp = s->rsb_list;
	s->rsb_list = temp->next;
	temp->next = r_list;
	r_list = temp;
	update_traffic_control(s, temp);
	if ((noisy_ & UPC_RESVTEAR) && (temp->timeout != -1)) {
	  Tcl& tcl = Tcl::instance();
	  tcl.evalf("%s upcall-resv-tear %d %d", name(),  s->sid, 
		    temp->sender->get_addr());
	}
      }
      if (s->rsb_list != NULL) {
	r = s->rsb_list;
	while (r != NULL) {
	  if ((r->next != NULL) &&
	      (r->next->sender->get_addr() == sender->get_addr()) &&
	      (r->next->nhop->get_hop() == nhop->get_hop())) {
	    temp = r->next;
	    r->next = temp->next;
	    temp->next = r_list;
	    r_list = temp;
	    num_rsb_--;
	    update_traffic_control(s, temp);
	    if ((noisy_ & UPC_RESVTEAR) && (temp->timeout != -1)) {
	      Tcl& tcl = Tcl::instance();
	      tcl.evalf("%s upcall-resv-tear %d %d", name(),  s->sid, 
			temp->sender->get_addr());
	    }
	  }
	  r = r->next;
	}
      }
      msg->delete_first(10);
      sender = (FILTER_SPEC *)msg->get_first(10);     
    }
  }
  if (r_list != NULL) {
    refresh_resv(s, 0);
    while ((r_list != NULL) && (find_rsb_for_psb(s, r_list->p))) {
      temp = r_list;
      r_list = r_list->next;
      clear_rsb(temp);
      delete temp;
    }
    if (r_list != NULL) {
      r = r_list;
      while (r->next != NULL) {
	if (find_rsb_for_psb(s, r->p)) {
	  temp = r->next;
	  r->next = temp->next;
	  clear_rsb(temp);
	  delete temp;
	}
      }
    }
    if (r_list != NULL) {
      send_resv_tear_messages(s, r_list);
    }
    while (r_list != NULL) {
      temp = r_list;
      r_list = r_list->next;
      clear_rsb(temp);
      delete temp;
    }
  }
}


void RSVPAgent::recv(Packet* p, Handler* h) {
  /* All packets are already being processed by the 'give' function,
     so get rid of this packet. */
  Packet::free(p);
}


/* Refresh the path/reservation state for all sessions that need to be
   refreshed (usually that should only be the first session in the 
   timer list).
   */
void RSVPAgent::refresh() {
  double now = Scheduler::instance().clock();
  /* PSB and RSB timeouts are checked with every refresh for this
     session and not when the actual timeout occurs. That means that
     path and reservation state might time out a bit slower in the
     simulation than in a real system (though no refresh messages from
     these states will be sent after the timeout), but this behaviour
     is not incorrect, since RFC 2205 states that L (a state's lifetime)
     should be equal *or greater* than (K + 0.5) * 1.5 * R. */
  if ((t_list_ != NULL) && in_range(minref(t_list_), now)) {
    check_path(t_list_);
    if (t_list_ != NULL) {
      check_resv(t_list_);
    }
    if (t_list_ != NULL) {
      if (in_range(minref(t_list_), now)) {
	if (minref(t_list_) == t_list_->path_ref) {
	  refresh_path(t_list_);
	} else {
	  /* Force a reservation refresh */
	  refresh_resv(t_list_, 1);
	}
      } else {
	reschedule_session(t_list_);
      }
    }
  }
}


/* Create and send path messages for this session, then reschedule
   the session. */
void RSVPAgent::refresh_path(session *s) {
  psb *p = s->psb_list;
  while (p != NULL) {
    if (!is_mcast(s->s->get_dest()) ||
	!is_leaf(p->sender->get_addr(), s->s->get_dest())) {
      send_path_message(s, p);
    }
    p = p->next;
  }
  s->path_ref = Scheduler::instance().clock() + s->path_tv->get_r();
  reschedule_session(s);
}


/* Create and send resv messages for this session, then reschedule
   the session. If the 'forced' flag is off, this function merely checks
   if a refresh is necessary (for example after a new RSB was added).
   If it finds a reservation state that has to be updated, it performs a
   normal refresh. */
void RSVPAgent::refresh_resv(session *s, char forced) {
  rsb *rsearch = s->rsb_list;
  rsb *rsearch2;
  while (rsearch != NULL) {
    if (!rsearch->processed) {
      /* Send a reservation message for all RSBs with the same PHOP
	 in their corresponding PSBs as rsearch. */
      if (send_resv_message(s, rsearch, forced) && !forced) {
	/* A reservation has changed and had to be forwarded. Force
	   a refresh for the whole reservation state. */
	forced = 1;
	rsearch2 = s->rsb_list;
	while (rsearch2 != NULL) {
	  rsearch2->processed = 0;
	  rsearch2 = rsearch2->next;
	}
	rsearch2 = s->rsb_list;
	while (rsearch2 != rsearch) {
	  if (rsearch->processed) {
	    send_resv_message(s, rsearch2, forced);
	  }
	  rsearch2 = rsearch2->next;
	}
      }
    }
    rsearch = rsearch->next;
  }
  rsearch = s->rsb_list;
  while (rsearch != NULL) {
    rsearch->processed = 0;
    rsearch = rsearch->next;
  }
  if (forced) {
    /* Only set a new refresh timeout if any messages were actually sent. */
    s->resv_ref = Scheduler::instance().clock() + s->resv_tv->get_r();
    reschedule_session(s);
  }
}


/* Release a session. Remove it from the timer list and from the
   session list, and send appropriate teardown messages for all path
   and reservation states. */
void RSVPAgent::release_session(session *s) {
  session *search, *temp;
  // Tcl& tcl = Tcl::instance(); - unused
  /* First remove the session from the timer list */
  if (t_list_ != NULL) {
    if (t_list_ == s) {
      t_list_ = t_list_->t_next;
    } else {
      search = t_list_;
      while ((search->t_next != s) && (search->t_next != NULL)) {
	search = search->t_next;
      }
      if (search->t_next != NULL) {
	search->t_next = search->t_next->t_next;
      }
    }
  }
  /* Now remove the session from the session list and send teardown
     messages for all existing path and reservation states. */
  if (s_list_ == s) {
    s_list_ = s_list_->next;
    temp = s;
  } else {
    search = s_list_;
    while ((search->next != s) && (search->next != NULL)) {
      search = search->next;
    }
    temp = search->next;
    search->next = search->next->next;
  }
  psb *p; 
  rsb *r;
  /* Tear down all reservation states */
  send_resv_tear_messages(s, s->rsb_list);
  while ((r = temp->rsb_list) != NULL) {
    temp->rsb_list = temp->rsb_list->next;
    update_traffic_control(temp, r);
    clear_rsb(r);
    delete r;
    num_rsb_--;
  }
  /* Tear down all path states */
  while ((p = temp->psb_list) != NULL) {
    send_path_tear_message(s, p);
    temp->psb_list = temp->psb_list->next;
    clear_psb(p);
    delete p;
    num_psb_--;
  }
  delete temp;
  if ((t_list_ != NULL) && (!t_list_->ref_flag)) {
    t_list_->ref_flag = 1;
    Tcl& tcl = Tcl::instance();
    tcl.evalf("[Simulator instance] at %f \"%s refresh\"",
	      minref(t_list_), name());
  }
}


/* Remove session from timer list and from the session list (unless
   it is a 'local' session. The difference between release_session and
   remove_session is that remove_session does not remove local sessions.
   */
void RSVPAgent::remove_session(session *s) {
  session *search;
  if (t_list_ == s) {
    t_list_ = t_list_->t_next;
  } else {
    search = t_list_;
    while (search->t_next != s) {
      search = search->t_next;
    }
    search->t_next = search->t_next->t_next;
  }
  if (!s->local) {
    release_session(s);
  }
  if ((t_list_ != NULL) && (!t_list_->ref_flag)) {
    t_list_->ref_flag = 1;
    Tcl& tcl = Tcl::instance();
    tcl.evalf("[Simulator instance] at %f \"%s refresh\"",
	      minref(t_list_), name());
  }
}


/* Reschedule a session */
void RSVPAgent::reschedule_session(session *s) {
  session *search;
  s->ref_flag = 0;
  if (t_list_ == NULL) {
    t_list_ = s;
  } else {
    if (t_list_ == s) {
      t_list_ = t_list_->t_next;
    } else {
      search = t_list_;
      /* A doubly linked list would make the following code faster,
	 but probably not simpler. However, that feature can be added
	 when it turns out that the simulations run too slow. */
      while ((search->t_next != NULL) && (search->t_next != s)) {
	search = search->t_next;
      }
      if (search->t_next == s) {
	search->t_next = search->t_next->t_next;
      }
    }
    if (t_list_ == NULL) {
      t_list_ = s;
    } else {
      if (minref(s) < minref(t_list_)) {
	s->t_next = t_list_;
	t_list_ = s;
      } else {
	search = t_list_;
	while ((search->t_next != NULL) && 
	       (minref(search->t_next) < minref(s))) {
	  search = search->t_next;
	}
	s->t_next = search->t_next;
	search->t_next = s; 
      }
    }
  }
  /* If the first element in the timer list hasn't been scheduled for
     refreshing yet, call Tcl to schedule it. In the ideal case, only
     the first element of the timer list is scheduled. The worst case, 
     that the scheduler has one entry for each session in each RSVP
     agent should be avoided.
     */
  if (!t_list_->ref_flag) {
    t_list_->ref_flag = 1;
    Tcl& tcl = Tcl::instance();
    tcl.evalf("[Simulator instance] at %f \"%s refresh\"",
	      minref(t_list_), name());
  }
}


/* Create and send a path message for a PSB in a session */
void RSVPAgent::send_path_message(session *s, psb *p) {
  if (nam_) {
    type_ = PT_RSVP_PATH;
  }
  // if ((s->s->get_dest() != addr_ >> Address::instance().NodeShift_[1]) &&
  if ((s->s->get_dest() != here_.addr_ ) && // SM
      (p->ttl >= 1)) {
    if (is_mcast(s->s->get_dest())) {
      dst_.addr_ = s->s->get_dest();
    } else {
      // dst_ = s->s->get_dest() << Address::instance().NodeShift_[1];
      dst_.addr_ = s->s->get_dest() ;
    }
    RSVPmessage *msg = new RSVPmessage(PATH, p->ttl-1);
    msg->add_object(new SESSION(s->s->get_contents()));
    msg->add_object(new TIME_VALUES(s->path_tv->get_contents()));
    msg->add_object(new SENDER_TEMPLATE(p->sender->get_contents()));
    msg->add_object(new SENDER_TSPEC(p->s_tspec->get_contents()));
    // msg->add_object(new RSVP_HOP(addr_ >> Address::instance().NodeShift_[1],
    msg->add_object(new RSVP_HOP(here_.addr_ , 
				 ip6_));
    if (ip6_) {
      size_ = msg->get_length() + 40; // Size for standard IPv6 header
    } else {
      size_ = msg->get_length() + 24;
    }
    fid_ = 46;
    Packet *pkt = allocpkt();
    hdr_cmn* hdr = hdr_cmn::access(pkt);
    pkt->allocdata(msg->get_conlength());
    memcpy(pkt->accessdata(), msg->get_contents(), msg->get_conlength());
    if (p->phop->get_hop() != -1) {
      hdr->iface_ = p->iif;
      hdr_ip *ip_hdr = hdr_ip::access(pkt);
      //ip_hdr->src_ = p->sender->get_addr() << Address::instance().NodeShift_[1];
      ip_hdr->saddr() = p->sender->get_addr() ; //SM
      p->check->give(pkt);
    } else {
      send(pkt, 0);
    }
    delete msg;
  }
}

/* Create and send a path tear message for a PSB in a session */
void RSVPAgent::send_path_tear_message(session *s, psb *p) {
  if (nam_) {
    type_ = PT_PATH_TEAR;
  }
  // if (((s->s->get_dest() != addr_ >> Address::instance().NodeShift_[1]) ||
  if (((s->s->get_dest() != here_.addr_ ) ||
       (s->s->get_dest() >= 1 << Address::instance().McastShift_)) &&
      (p->ttl > 1)) {
    if (!is_mcast(s->s->get_dest()) ||
	!is_leaf(p->sender->get_addr(), s->s->get_dest())) {
      if (s->s->get_dest() >= 1 << Address::instance().McastShift_) {
	dst_.addr_ = s->s->get_dest();
      } else {
	// dst_ = s->s->get_dest() << Address::instance().NodeShift_[1];
	dst_.addr_ = s->s->get_dest() ;
      }
      RSVPmessage *msg = new RSVPmessage(PATHTEAR, p->ttl-1);
      msg->add_object(new SESSION(s->s->get_contents()));
      msg->add_object(new SENDER_TEMPLATE(p->sender->get_contents()));
      //msg->add_object(new RSVP_HOP(addr_ >> Address::instance().NodeShift_[1],
      msg->add_object(new RSVP_HOP(here_.addr_ , 
				   ip6_));
      /* A SENDER_TSPEC object may be added to the message here. RFC 2205
	 is unclear about that though. At one point it says that
	 a SENDER_TSPEC is an obligatory part of a sender descriptor,
	 later it says that it may be ommited in a PathTear message and
	 will be ignored. */
      size_ = msg->get_length();
      if (ip6_) {
	size_ = msg->get_length() + 40; // Size for standard IPv6 header
      } else {
	size_ = msg->get_length() + 24;
      }
      fid_ = 46;
      Packet *pkt = allocpkt();
      hdr_cmn* hdr = hdr_cmn::access(pkt);
      pkt->allocdata(msg->get_conlength());
      memcpy(pkt->accessdata(), msg->get_contents(), msg->get_conlength());
      if (p->phop->get_hop() != -1) {
	hdr->iface_ = p->iif;
	hdr_ip *ip_hdr = hdr_ip::access(pkt);
	// ip_hdr->src_ = p->sender->get_addr() << Address::instance().NodeShift_[1];
	ip_hdr->saddr() = p->sender->get_addr() ; //SM
	
	p->check->give(pkt);
      } else {
	send(pkt, 0);
      }
      delete msg;
    }
  }
}


/* Create and send a resv confirm message. */
void RSVPAgent::send_resv_conf_message(session *s, rsb *r) {
  if (nam_) {
    type_ = PT_RESV_CONF;
  }
  RSVPmessage *msg = new RSVPmessage(RESVCONF, 32);
  msg->add_object(new SESSION(s->s->get_contents()));
  msg->add_object(new STYLE(r->st->get_contents()));
  msg->add_object(new FILTER_SPEC(r->sender->get_addr(), ip6_));
  // msg->add_object(new ERROR_SPEC(addr_ >> Address::instance().NodeShift_[1],
  msg->add_object(new ERROR_SPEC(here_.addr_ , 
				 0, 0, 0, ip6_));
  msg->add_object(new RESV_CONFIRM(r->confirm->get_contents()));
  if (ip6_) {
    size_ = msg->get_length() + 40; // Size for standard IPv6 header
  } else {
    size_ = msg->get_length() + 24;
  }
  Tcl& tcl = Tcl::instance();
  tcl.evalf("[Simulator instance] connect-rsvp-agents %d %d", 
	    // addr_ >> Address::instance().NodeShift_[1],
	    here_.addr_ ,
	    r->confirm->get_addr());
  fid_ = 46;
  Packet *pkt = allocpkt();
  pkt->allocdata(msg->get_conlength());
  memcpy(pkt->accessdata(), msg->get_contents(), msg->get_conlength());
  send(pkt, 0);
  delete msg;
}


/* Create and send a resv message based on a reservation message that
   was received. */
void RSVPAgent::send_resv_err_message(RSVPmessage *resvmsg, int code) {
  if (nam_) {
    type_ = PT_RESV_ERR;
  }
  RSVPmessage *msg = new RSVPmessage(RESVERR, 32);
  msg->add_object(new SESSION(((SESSION *)resvmsg->get_first(1))->get_contents()));
  // msg->add_object(new RSVP_HOP(addr_ >> Address::instance().NodeShift_[1],
  msg->add_object(new RSVP_HOP(here_.addr_ , 
			       ip6_));
  msg->add_object(new STYLE(((STYLE *)resvmsg->get_first(8))->get_contents()));
  FILTER_SPEC *sender;
  sender = (FILTER_SPEC *)resvmsg->get_first(10);
  while (sender != NULL) {
    msg->add_object(new FILTER_SPEC(sender->get_contents()));
    resvmsg->delete_first(10);
    sender = (FILTER_SPEC *)resvmsg->get_first(10);
  }
  // msg->add_object(new ERROR_SPEC(addr_ >> Address::instance().NodeShift_[1],
  msg->add_object(new ERROR_SPEC(here_.addr_ , 
				 0, code, 0, ip6_));
  if (ip6_) {
    size_ = msg->get_length() + 40; // Size for standard IPv6 header
  } else {
    size_ = msg->get_length() + 24;
  }
  // dst_ = ((RSVP_HOP *)resvmsg->get_first(3))->get_hop() << Address::instance().NodeShift_[1];
  dst_.addr_ = ((RSVP_HOP *)resvmsg->get_first(3))->get_hop() ; 
  fid_ = 46;
  Packet *pkt = allocpkt();
  pkt->allocdata(msg->get_conlength());
  memcpy(pkt->accessdata(), msg->get_contents(), msg->get_conlength());
  send(pkt, 0);
  delete msg;
}


/* Create and send a resv err message (an admission control error,
   to be exact) for an RSB. */
void RSVPAgent::send_resv_err_message(session *s, rsb *r) {
  if (nam_) {
    type_ = PT_RESV_ERR;
  }
  RSVPmessage *msg = new RSVPmessage(RESVERR, 32);
  msg->add_object(new SESSION(s->s->get_contents()));
  // msg->add_object(new RSVP_HOP(addr_ >> Address::instance().NodeShift_[1],
  msg->add_object(new RSVP_HOP(here_.addr_ , 
			       ip6_));
  msg->add_object(new STYLE(r->st->get_contents()));
  msg->add_object(new FILTER_SPEC(r->sender->get_addr(), ip6_));
  // msg->add_object(new ERROR_SPEC(addr_ >> Address::instance().NodeShift_[1],
  msg->add_object(new ERROR_SPEC(here_.addr_ , 
				 0, 1, 0, ip6_));
  if (ip6_) {
    size_ = msg->get_length() + 40; // Size for standard IPv6 header
  } else {
    size_ = msg->get_length() + 24;
  }
  // dst_ = r->nhop->get_hop() << Address::instance().NodeShift_[1];
  dst_.addr_ = r->nhop->get_hop() ; 
  fid_ = 46;
  Packet *pkt = allocpkt();
  pkt->allocdata(msg->get_conlength());
  memcpy(pkt->accessdata(), msg->get_contents(), msg->get_conlength());
  send(pkt, 0);
  delete msg;
}


/* Create and send a resv err message for all RSBs with the same PHOP in
   their corresponding PSBs. */
void RSVPAgent::send_resv_err_message(session *s, 
				      RSVPmessage *resvmsg, rsb *r) {
  if (nam_) {
    type_ = PT_RESV_ERR;
  }
  if (r->nhop->get_hop() != -1) {
    RSVPmessage *msg = new RSVPmessage(RESVERR, 32);
    msg->add_object(new SESSION(((SESSION *)resvmsg->get_first(1))->get_contents()));
    // msg->add_object(new RSVP_HOP(addr_ >> Address::instance().NodeShift_[1],
    msg->add_object(new RSVP_HOP(here_.addr_ , 
				 ip6_));
    msg->add_object(new STYLE(((STYLE *)resvmsg->get_first(8))->get_contents()));
    rsb *rsearch = r;
    while (rsearch != NULL) {
      if (rsearch->processed && 
	  (rsearch->nhop->get_hop() == r->nhop->get_hop())) {
	rsearch->processed = 0;
	msg->add_object(new FILTER_SPEC(rsearch->sender->get_addr(), ip6_));
      }
      rsearch = rsearch->next;
    }
    msg->add_object(new ERROR_SPEC(((ERROR_SPEC *)resvmsg->get_first(6))->get_contents()));
    if (ip6_) {
      size_ = msg->get_length() + 40; // Size for standard IPv6 header
    } else {
      size_ = msg->get_length() + 24;
    }
    // dst_ = r->nhop->get_hop() << Address::instance().NodeShift_[1];
    dst_.addr_ = r->nhop->get_hop() ; 
    fid_ = 46;
    Packet *pkt = allocpkt();
    pkt->allocdata(msg->get_conlength());
    memcpy(pkt->accessdata(), msg->get_contents(), msg->get_conlength());
    send(pkt, 0);
    delete msg;
  }
}


/* Create and send a resv message for all RSBs with the same PHOP in
   their corresponding PSBs. */
int RSVPAgent::send_resv_message(session *s, rsb *r, char forced) {
  if (nam_) {
    type_ = PT_RSVP_RESV;
  }
  char snd = 0;
  if (r->p->phop->get_hop() != -1) {
    RSVPmessage *msg = new RSVPmessage(RESV, 32);
    msg->add_object(new SESSION(s->s->get_contents()));
    msg->add_object(new TIME_VALUES(s->resv_tv->get_contents()));
    // msg->add_object(new RSVP_HOP(addr_ >> Address::instance().NodeShift_[1],
    msg->add_object(new RSVP_HOP(here_.addr_ , 
				 ip6_));
    msg->add_object(new STYLE(r->st->get_contents()));
    if (s->confirm) {
      // msg->add_object(new RESV_CONFIRM(addr_ >> Address::instance().NodeShift_[1], ip6_));
      msg->add_object(new RESV_CONFIRM(here_.addr_ , ip6_));
      s->confirm = 0;
    }
    rsb *rsearch = s->rsb_list;
    nsaddr_t phop = r->p->phop->get_hop();
    nsaddr_t sender;
    double lub_rate;
    long lub_size;
    rsb *rsearch2, *rsearch3;
    char merged;
    tcsb *t;
    while (rsearch != NULL) {
      if (!rsearch->processed && (rsearch->p->phop->get_hop() == phop)) {
	/* Calculate the LUB for all RSBs that match 'sender'. */
	lub_rate = 0;
	lub_size = 0;
	sender = rsearch->sender->get_addr();
	rsearch2 = rsearch;
	while (rsearch2 != NULL) {
	  if (rsearch2->confirm != NULL) {
	    merged = 0;
	    rsearch3 = s->rsb_list;
	    while (!merged && (rsearch3 != NULL)) {
	      if ((rsearch3->p->phop->get_hop() == phop) &&
		  (rsearch3 != rsearch2) &&
		  (rsearch2->fspec->get_rate() <= 
		   rsearch3->fspec->get_rate()) && 
		  (rsearch2->fspec->get_size() <=
		   rsearch3->fspec->get_rate())) {
		merged = 1;
	      }
	      rsearch3 = rsearch3->next;
	    }
	    if (merged) {
	      send_resv_conf_message(s, rsearch2);
	    } else {
	      msg->add_object(new RESV_CONFIRM(rsearch2->confirm->get_contents()));
	    }
	    delete rsearch2->confirm;
	    rsearch2->confirm = NULL;
	  }
	  if ((rsearch2->p->phop->get_hop() == phop) &&
	      (rsearch2->sender->get_addr() == sender)) {
	    if (rsearch2->fspec->get_rate() > lub_rate) {
	      lub_rate = rsearch2->fspec->get_rate();
	    }
	    if (rsearch2->fspec->get_size() > lub_size) {
	      lub_size = rsearch2->fspec->get_size();
	    }
	    rsearch2->processed = 1;
	  }
	  rsearch2 = rsearch2->next;
	}
	t = s->tcsb_list;
	while ((t != NULL) && ((t->sender != sender) || 
			       (t->oifhop != rsearch->oifhop))) {
	  t = t->next;
	}
	if (t == NULL) {
	  t = new tcsb();
	  t->fl = new FLOWSPEC(0, 0);
	  t->ofl = new FLOWSPEC(0, 0);
	  t->oifhop = rsearch->oifhop;
	  t->sender = sender;
	  t->next = s->tcsb_list;
	  s->tcsb_list = t;
	}
	if ((lub_rate != t->ofl->get_rate()) || 
	    (lub_size != t->ofl->get_size()) ||
	    forced) {
	  msg->add_object(new FLOWSPEC(lub_rate, lub_size));
	  msg->add_object(new FILTER_SPEC(sender, ip6_));
	  snd = 1;
	  delete t->ofl;
	  t->ofl = new FLOWSPEC(lub_rate, lub_size);
	}
      }
      rsearch = rsearch->next;
    }
    if (snd) {
      if (ip6_) {
	size_ = msg->get_length() + 40; // Size for standard IPv6 header
      } else {
	size_ = msg->get_length() + 24;
      }
      // dst_ = r->p->phop->get_hop() << Address::instance().NodeShift_[1];
      dst_.addr_ = r->p->phop->get_hop() ;
      fid_ = 46;
      Packet *pkt = allocpkt();
      pkt->allocdata(msg->get_conlength());
      memcpy(pkt->accessdata(), msg->get_contents(), msg->get_conlength());
      send(pkt, 0);
      delete msg;
    }
  }
  return snd;
}


/* Create and send a resv tear message for a list of RSBs in a session */
void RSVPAgent::send_resv_tear_messages(session *s, rsb *r_list) {
  if (nam_) {
    type_ = PT_RESV_TEAR;
  }
  rsb *rsearch = r_list;
  rsb *rsearch2, *rsearch3;
  while (rsearch != NULL) {
    if ((rsearch->p->phop->get_hop() != -1) && !rsearch->processed) {
      RSVPmessage *msg = new RSVPmessage(RESVTEAR, 32);
      msg->add_object(new SESSION(s->s->get_contents()));
      // msg->add_object(new RSVP_HOP(addr_ >> Address::instance().NodeShift_[1],
      msg->add_object(new RSVP_HOP(here_.addr_ , 
				   ip6_));
      msg->add_object(new STYLE(rsearch->st->get_contents()));
      nsaddr_t phop = rsearch->p->phop->get_hop();
      nsaddr_t sender;
      rsearch2 = rsearch;
      while (rsearch2 != NULL) {
	sender = rsearch2->sender->get_addr();
	if (!rsearch2->processed && (rsearch2->p->phop->get_hop() == phop)) 
	  {
	    sender = rsearch2->sender->get_addr();
	    rsearch3 = rsearch;
	    while (rsearch3 != NULL) {
	      if ((rsearch3->p->phop->get_hop() == phop) &&
		  (rsearch3->sender->get_addr() == sender)) {
		/* The 'processed' flags will not have to be deleted from
		   the RSBs later, since the r_list will always be deleted
		   after being processed by this function. */
		rsearch3->processed = 1;
	      }
	      rsearch3 = rsearch3->next;
	  }
	  msg->add_object(new FILTER_SPEC(sender, ip6_));
	}
	rsearch2 = rsearch2->next;
      }
      if (ip6_) {
	size_ = msg->get_length() + 40; // Size for standard IPv6 header
      } else {
	size_ = msg->get_length() + 24;
      }
      // dst_ = rsearch->p->phop->get_hop() << Address::instance().NodeShift_[1];
      dst_.addr_ = rsearch->p->phop->get_hop() ;
      fid_ = 46;
      Packet *pkt = allocpkt();
      // hdr_cmn* hdr = (hdr_cmn*)pkt->access(off_cmn_); - unused
      pkt->allocdata(msg->get_conlength());
      memcpy(pkt->accessdata(), msg->get_contents(), msg->get_conlength());
      send(pkt, 0);
      delete msg;
    }
    rsearch = rsearch->next;
  }
}


/* Update an existing PSB with new values. */
void RSVPAgent::update_psb(session *s, psb *p, double rate, long bucket, 
			   char ttl, nsaddr_t sender, nsaddr_t phop, 
			   double refresh, int iface, RSVPChecker *check) {
  /* The following is slightly different from what the RSVP Processing
     Rules suggest for updated PSBs */
  char diff = ((p->sender->get_addr() != sender) || 
	       (p->s_tspec->get_rate() != rate) ||
	       (p->s_tspec->get_size() != bucket) ||
	       (p->ttl != ttl));
  clear_psb(p);
  p->sender = new SENDER_TEMPLATE(sender, ip6_);
  p->s_tspec = new SENDER_TSPEC(rate, bucket);
  p->phop = new RSVP_HOP(phop, ip6_);
  p->ttl = ttl;
  p->iif = iface;
  p->check = check;
  if (phop > -1) { // Doesn't make sense when call came from API
    p->timeout = (lifetime_factor_ + 0.5) * 1.5 * refresh + 
      Scheduler::instance().clock();
  }
  /* Send a path message for this PSB to inform downstream nodes
     of the updated path state. */
  if ((diff) || 
      ((is_mcast(s->s->get_dest())) && 
	(is_leaf(p->sender->get_addr(), s->s->get_dest())))) {
    send_path_message(s, p);
    s->path_ref = Scheduler::instance().clock() + s->path_tv->get_r();
    reschedule_session(s);
    /* Make an upcall to the API if necessary */
    if ((noisy_ & UPC_PATH) && (diff)) {
      Tcl& tcl = Tcl::instance();
      tcl.evalf("%s upcall-path %d %f %d %d", name(),  s->sid, rate, 
		bucket, sender);
    }
  }
}


/* Update an existing RSB with new values. */
void RSVPAgent::update_rsb(session *s, rsb *r, psb *p, nsaddr_t sender, 
			   FLOWSPEC *fl, const char *style, nsaddr_t nhop, 
			   double refresh, nsaddr_t fromhop) {
  long nstyle;
  if (strcasecmp(style, "FF") == 0) {
    nstyle = STYLE_FF;
  }
  r->modified = ((r->fspec->get_rate() != fl->get_rate()) ||
		 (r->fspec->get_size() != fl->get_size()) ||
		 (r->st->get_style() != nstyle) ||
		 (r->p != p) ||
		 (r->oifhop != fromhop));
  if (nhop > -1) { // Doesn't make sense when call came from API
    r->timeout = (lifetime_factor_ + 0.5) * 1.5 * refresh + 
      Scheduler::instance().clock();
  }
  /* Store the old values, so in case the reservation is rejected by
     admission control, it can be restored. */
  r->old = new rsb();
  r->old->nhop = r->nhop;
  r->old->sender = r->sender;
  r->old->fspec = r->fspec;
  r->old->st = r->st;
  r->old->oifhop = r->oifhop;
  r->old->p = r->p;
  if (r->modified) {
    r->nhop = new RSVP_HOP(nhop, ip6_);
    r->sender = new FILTER_SPEC(sender, ip6_);
    r->fspec = new FLOWSPEC(fl->get_contents());
    r->st = new STYLE(nstyle);
    r->oifhop = fromhop;
    r->p = p;
  }
}



/* The following function has not been tested thoroughly for
   measurement based admission control algorithms. It will probably be
   replaced sooner or later by algorithm-specific functions. */
int modify_flow(ADC *adc, int cl, double r_old, int b_old, double r, int b) {
  if (adc->peak_rate(cl, r, b) <= adc->peak_rate(cl, r_old, b_old)) {
    /* The new reservation is smaller than the old one =>
       notify admission control of the change and return 1 */
    adc->teardown_action(cl, r_old, b_old);
    adc->admit_flow(cl, r, b);
    return 1;
  } else {
    /* Check if the new flow can be admitted. */
    adc->teardown_action(cl, r_old, b_old);
    if (adc->admit_flow(cl, r, b)) {
      /* Yes => return 1 */
      return 1;
    } else {
      /* No => re-install the old flow */
      adc->admit_flow(cl, r_old, b_old);
      return 0;
    }
  }
}



/* Find all RSBs that match the (session, sender, nhop) triple for the
   'active' RSB, compute the LUB of their flowspecs then query admission 
   control and perform a reservation if possible. This function only works 
   for the FF style. */
int RSVPAgent::update_traffic_control(session *s, rsb *r) {
  rsb *rsearch = s->rsb_list;
  nsaddr_t sender = r->sender->get_addr();
  nsaddr_t nhop = r->nhop->get_hop();
  nsaddr_t oifhop = r->oifhop;
  double lub_rate = 0;
  long lub_size = 0;
  tcsb *t = s->tcsb_list;
  tcsb *ts;
  ADC *adc;
  char admitted;
  char refresh_needed = 0;
  char is_biggest = 1;
  Tcl& tcl = Tcl::instance();
  if (nhop != -1) {
    /* First find the matching TCSB for the (session, sender, oif) triple */
    while ((t != NULL) && ((t->sender != sender) || (t->oifhop != oifhop))) {
      t = t->next;
    }
    /* Now compute the LUB for the flowspecs from all matching RSBs. */
    while (rsearch != NULL) {
      if ((rsearch->sender->get_addr() == sender) && 
	  (rsearch->oifhop == oifhop)) {
	if ((r != rsearch) &&
	    (rsearch->fspec->get_rate() >= r->fspec->get_rate()) &&
	    (rsearch->fspec->get_size() >= r->fspec->get_size())) {
	  is_biggest = 0;
	}
	if (rsearch->fspec->get_rate() > lub_rate) {
	  lub_rate = rsearch->fspec->get_rate();
	}
	if (rsearch->fspec->get_size() > lub_size) {
	  lub_size = rsearch->fspec->get_size();
	}
      }
      rsearch = rsearch->next;
    }
    /* Check first if the LUB flowspec is zero and a tcspec exists.
       In that case, remove the existing reservation, because this
       can only happen when an RSB was removed (e.g. when a teardown
       message arrived). */
    if ((t != NULL) && (lub_rate == 0)) {
      /* Notify admission control of the change. */
      tcl.evalf("%s get-adc %d %d", name(),
		// addr_ >> Address::instance().NodeShift_[1], r->oifhop);
		here_.addr_ , r->oifhop);
      adc = (ADC *)TclObject::lookup(tcl.result());
      modify_flow(adc, 1, t->fl->get_rate(), t->fl->get_size(), 0, 0);
      tcl.evalf("[Simulator instance] delete-filter-fid %d %d %d %d",
		// addr_ >> Address::instance().NodeShift_[1], r->oifhop,
		here_.addr_ , r->oifhop,
		sender, s->s->get_fid());
      num_flows_--;
      tcl.evalf("[Simulator instance] delete-reservation %d %d %s",
		// addr_ >> Address::instance().NodeShift_[1], r->oifhop,
		here_.addr_ , r->oifhop,
		t->handle->name());
      /* Delete t */
      if (t == s->tcsb_list) {
	s->tcsb_list = t->next;
      } else {
	ts = s->tcsb_list;
	while (ts->next != t) {
	  ts = ts->next;
	}
	ts->next = t->next;
      }
      delete t->fl;
      delete t->ofl;
      delete t;
    } else {
      /* Now query admission control for lub_rate and lub_size. */
      tcl.evalf("%s get-adc %d %d", name(),
		// addr_ >> Address::instance().NodeShift_[1], r->oifhop);
		here_.addr_ , r->oifhop);
      adc = (ADC *)TclObject::lookup(tcl.result());
      if (t == NULL) { // New flow
	admitted = adc->admit_flow(1, lub_rate, lub_size);
      } else { // Modification of existing flow
	admitted = modify_flow(adc, 1, t->fl->get_rate(),
				    t->fl->get_size(), lub_rate, lub_size);
      }
      if (!admitted) {
	send_resv_err_message(s, r);
	if (r->confirm != NULL) {
	  delete r->confirm;
	  r->confirm = NULL;
	}
	/* Delete the active rsb if it is new, or return it to the old state
	   if it was modified. */
	if (r->is_new) {
	  if (s->rsb_list == r) {
	    s->rsb_list = r->next;
	    delete r;
	  } else {
	    rsearch = s->rsb_list;
	    while ((rsearch != NULL) && (rsearch->next != r)) {
	      rsearch = rsearch->next;
	    }
	    rsearch->next = r->next;
	    clear_rsb(r);
	    delete r;
	    num_rsb_--;
	  }
	} else {
	  clear_rsb(r);
	  r->nhop = r->old->nhop;
	  r->sender = r->old->sender;
	  r->fspec = r->old->fspec;
	  r->st = r->old->st;
	  r->p = r->old->p;
	  r->oifhop = r->old->oifhop;
	  delete r->old;
	  r->old = NULL;
	}
      } else {
	/* Delete all the pointers to the former RSB values in the
	   active RSB if it was modified. They are no longer needed. */
	if (!r->is_new) {
	  delete r->old;
	  r->old = NULL;
	}
	/* Does the RSB have a RESV_CONFIRM object and was it merged into
	   an existing reservation, or is this the leaf node for this RSB? 
	   If so, send a ResvConf message and remove the RESV_CONFIRM 
	   object. */
	if ((r->confirm != NULL) && 
	    (!is_biggest || (r->p->phop->get_hop() == -1))) {
	  send_resv_conf_message(s, r);
	  delete r->confirm;
	  r->confirm = NULL;
	}
	/* Perform the actual reservation. If there is no TCSB yet, 
	   create one and get a new handle to a WFQClass from Tcl. Otherwise
	   modify the reservation with the existing handle. */
	if (t == NULL) {
	  t = new tcsb();
	  t->fl = new FLOWSPEC(lub_rate, lub_size);
	  t->ofl = new FLOWSPEC(0, 0);
	  t->oifhop = oifhop;
	  t->sender = sender;
	  t->next = s->tcsb_list;
	  s->tcsb_list = t;
	  tcl.evalf("[Simulator instance] add-reservation %d %d %f %d",
		    // addr_ >> Address::instance().NodeShift_[1],
		    here_.addr_ ,
		    r->oifhop, lub_rate, lub_size);
	  num_flows_++;
	  t->handle = TclObject::lookup(tcl.result());
	  tcl.evalf("[Simulator instance] add-filter-fid %d %d %s %d %d",
		    // addr_ >> Address::instance().NodeShift_[1], r->oifhop,
		    here_.addr_ , r->oifhop,
		    t->handle->name(), sender, s->s->get_fid());
	  refresh_needed = 1;
	} else {
	  if ((t->fl->get_rate() != lub_rate) || (t->fl->get_size() != lub_size)) {
	    delete t->fl;
	    t->fl = new FLOWSPEC(lub_rate, lub_size);
	    tcl.evalf("[Simulator instance] modify-reservation %d %d %s %f %d",
		      // addr_ >> Address::instance().NodeShift_[1],
		      here_.addr_ ,
		      r->oifhop, t->handle->name(), lub_rate, lub_size);
	    refresh_needed = 1;
	  }
	}
      }
    }
  }
  return refresh_needed;
}
